package lightsout;

import java.util.ArrayList;

/** Represents a board of Lights out. */
public class Board {
	
	public static final int OFF  = 0;
	public static final int ON = 1;
	public static final int OFF_AND_PRESSED = 2;
	public static final int ON_AND_PRESSED = 3;
	
	/** Checks whether the state is pressed. */
	public static boolean isPressed(int val) {
		return val == OFF_AND_PRESSED || val == ON_AND_PRESSED;
	}
	
	/** Checks whether the state is ON. */
	public static boolean isOn(int val) {
		return val == ON || val == ON_AND_PRESSED;
	}
	
	/** Checks whether the state is OFF. */
	public static boolean isOff(int val) {
		return val == OFF || val == OFF_AND_PRESSED;
	}

	/* Cells of the board. */
	private final ArrayList<ArrayList<Integer>> states = new ArrayList<>();
	
	public Board(int N, int M) {
		for (int i = 0; i < N; ++i) {
			states.add(new ArrayList<>());
			for (int j = 0; j < M; ++j) {
				states.get(i).add(OFF);
			}
		}
	}
	
	public Board(Board board) {
		for (ArrayList<Integer> row : board.states) {
			states.add(new ArrayList<>(row));
		}
	}
	
	/** Returns the number of rows in the board. */
	public int getRows() {
		return states.size();
	}
	
	/** Returns the number of cols in the board. */
	public int getCols() {
		return states.get(0).size();
	}
	
	/**
	 * Set the value of the board to a given state.
	 */
	public void set(int x, int y, int to) {
		states.get(x).set(y,  to);
	}
	
	/** Checks whether the given position was pressed. */
	public boolean isPressed(int x, int y) {
		return isPressed(states.get(x).get(y));
	}
	
	/** Checks whether the given board is solved. */
	public boolean isSolved() {
		for (ArrayList<Integer> row : states) {
			for (Integer val : row) {
				if (isOn(val)) {
					return false;
				}
			}
		}
		return true;
	}
	
	/**
	 * Returns whether the given coordinates represent a valid position
	 * on the board.
	 */
	private boolean isValid(int x, int y) {
		return x >= 0 && x < states.size() && y >= 0 && y < states.size();
	}
	
	/**
	 * Press the light button at (x, y). This will change the state of itself
	 * and neighbours.
	 * 
	 * Note: You are not allowed (and it makes no sense) to toggle the same state
	 * twice.
	 */
	public void press(int x, int y) {
		if (!isValid(x, y)) {
			throw new IndexOutOfBoundsException("Invalid position in the board.");
		}
		if (isPressed(states.get(x).get(y))) {
			throw new IllegalArgumentException("This state was already pressed.");
		}
		// Toggle the state of the 5 lights.
		for (int dx = -1; dx <= 1; ++dx) {
			for (int dy = -1; dy <= 1; ++dy) {
				if (dx != 0 && dy != 0) continue;
				if (!isValid(x + dx, y + dy)) continue;
				
				int prev = states.get(x + dx).get(y + dy);
				states.get(x + dx).set(y + dy, (prev ^ 1));
			}
		}
		// Make (x, y) pressed.
		states.get(x).set(y, states.get(x).get(y) + 2);
	}
	
	/** Undoes a press. Useful for backtracking. */
	public void undoPress(int x, int y) {
		if (!isValid(x, y)) {
			throw new IndexOutOfBoundsException("Invalid position in the board.");
		}
		if (!isPressed(states.get(x).get(y))) {
			throw new IllegalArgumentException("This state has not been pressed.");
		}
		// Toggle the state of the 5 lights.
		for (int dx = -1; dx <= 1; ++dx) {
			for (int dy = -1; dy <= 1; ++dy) {
				if (dx != 0 && dy != 0) continue;
				if (!isValid(x + dx, y + dy)) continue;
				
				int prev = states.get(x + dx).get(y + dy);
				states.get(x + dx).set(y + dy, (prev ^ 1));
			}
		}
		// Make (x, y) pressed.
		states.get(x).set(y, states.get(x).get(y) - 2);
	}
	
	
	@Override
	public String toString() {
		StringBuilder str = new StringBuilder();
		for (ArrayList<Integer> row : states) {
			for (Integer cell : row) {
				str.append(cell & 1);
				str.append(' ');
			}
			str.append('\n');
		}
		return str.toString();
	}
	
}
